from django.db import models
from ua_parser import user_agent_parser
from solo.models import SingletonModel


class AccessLog(models.Model):
    timestamp = models.DateTimeField(db_index=True)
    bytes_sent = models.CharField(max_length=500)
    referer = models.TextField()
    request = models.TextField()
    http_method = models.CharField(max_length=500)
    http_version = models.CharField(max_length=500)
    remote_host = models.CharField(max_length=500)
    remote_login = models.CharField(max_length=500)
    remote_user = models.CharField(max_length=500)
    status = models.CharField(max_length=500)
    user_agent = models.CharField(max_length=500)
    request_time = models.CharField(max_length=500)
    upstream_response_time = models.CharField(max_length=500)
    pipe = models.CharField(max_length=500)

    class Meta:
        ordering = ('-timestamp',)

    def user_agent_display(self):
        info = user_agent_parser.Parse(self.user_agent)
        return "%s %s %s" % (info['os']['family'] or '',
                             info['user_agent']['family'] or '',
                             info['user_agent']['major'] or '')

    user_agent_display.short_description = 'User agent'


BOT_LIST = ['cloudflare', 'twiceler', 'yahooseeker', 'chtml', 'generic', 'heritrix', 'attentio', 'fast',
            'mediapartners', 'python', 'experiment', 'fastmobilecrawl', 'curl',
            'yahooysmcm', 'crawl', 'bingbot', 'bot', 'borg', 'google(^tv)', 'yahoo', 'slurp', 'msnbot', 'msrbot',
            'openbot', 'archiver', 'netresearch', 'lycos', 'scooter',
            'altavista', 'teoma', 'gigabot', 'baiduspider', 'blitzbot', 'oegp', 'charlotte', 'furlbot', 'http%20client',
            'polybot', 'htdig', 'ichiro', 'mogimogi', 'larbin',
            'pompos', 'scrubby', 'searchsight', 'seekbot', 'semanticdiscovery', 'silk', 'snappy', 'speedy', 'spider',
            'voila', 'vortex', 'voyager', 'zao', 'zeal',
            'fast\-webcrawler', 'converacrawler', 'dataparksearch', 'findlinks', 'crawler', 'Netvibes',
            'Sogou Pic Spider', 'ICC\-Crawler', 'Innovazion Crawler', 'Daumoa',
            'EtaoSpider', 'A6\-Indexer', 'YisouSpider', 'Riddler', 'DBot', 'wsr\-agent', 'Xenu', 'SeznamBot',
            'PaperLiBot', 'SputnikBot', 'CCBot', 'ProoXiBot', 'Scrapy',
            'Genieo', 'Screaming Frog', 'YahooCacheSystem', 'CiBra', 'Nutch', 'holmes', 'WebThumbnail', '^voyager',
            'heritrix', 'scraper', 'favicon', 'Google.*/\\+/web/snippet',
            'Icarus6j', 'PagePeeker', '^vortex', 'Sogou', '^Java/', 'BlogBridge', 'ZooShot', 'indexer', 'GomezAgent',
            '^JNLP/', '^NING', 'WinHTTP', 'TLSProber', 'Squrl Java',
            'NewsGator', 'Google-HTTP-Java-Client', 'Reaper', 'WhatWeb', 'crawl', 'facebookexternalhit',
            'Python-urllib', 'IlTrovatore-Setaccio', 'AppEngine-Google', 'GoogleStackdriverMonitoring',
            'InternetArchive', 'WordPress', 'Retreiver']


class AccessLogConfiguration(SingletonModel):
    user_agent_bot_list = models.TextField(default=', '.join(BOT_LIST))
    access_log_export_excluded_ips = models.TextField(default='')

    def __unicode__(self):
        return u"Access Log Configuration"

    class Meta:
        verbose_name = "Access Log Configuration"



import datetime

import import_export
import pytz
from django import forms
from django.conf import settings
from django.contrib import admin
from django.contrib.admin import SimpleListFilter
from django.db.models import Q
from django.http import HttpRequest, HttpResponse
from django.template.response import TemplateResponse
from import_export import admin as import_export_admin
from import_export import fields, resources
from import_export import forms as import_export_forms
from solo.admin import SingletonModelAdmin

from .models import AccessLog, AccessLogConfiguration


def make_tz_aware(dt, tz='UTC', is_dst=None):
    """Add timezone information to a datetime object, only if it is naive."""
    tz = dt.tzinfo or tz
    try:
        tz = pytz.timezone(tz)
    except AttributeError:
        pass
    return tz.localize(dt, is_dst=is_dst)


class AccessLogResource(resources.ModelResource):
    user_agent_raw = fields.Field()

    class Meta:
        model = AccessLog
        export_order = ('id', 'timestamp', 'bytes_sent', 'referer', 'request', 'http_method', 'http_version',
                        'remote_host', 'remote_login', 'remote_user', 'status', 'user_agent', 'user_agent_raw',
                        'request_time', 'upstream_response_time', 'pipe')

    def dehydrate_timestamp(self, access_log):
        return make_tz_aware(access_log.timestamp, settings.TIME_ZONE).strftime("%m/%d/%y %H:%M:%S")

    def dehydrate_user_agent(self, access_log):
        return access_log.user_agent_display()

    def dehydrate_user_agent_raw(self, access_log):
        return access_log.user_agent


class CustomExportForm(import_export_forms.ExportForm):
    file_format = forms.ChoiceField(label='Format*', choices=())
    start_date = forms.DateField(label='Start Date', required=False, widget=admin.widgets.AdminDateWidget())
    end_date = forms.DateField(label='End Date', required=False, widget=admin.widgets.AdminDateWidget())
    status = forms.ChoiceField(label='Status', required=False, choices=())
    filter_bots = forms.BooleanField(label="Filter Bots", required=False, initial=True)
    filter_admin_ips = forms.ChoiceField(label="Filter Admin IPs?", required=False, initial=90,
                                         choices=((90, 'From last 90 days'), (60, 'From the last 60 days'),
                                                  (30, 'From last 30 days'), ('', 'Don\'t Filter')))

    def __init__(self, formats, *args, **kwargs):
        super(CustomExportForm, self).__init__(formats, *args, **kwargs)
        status_list = AccessLog.objects.order_by().values_list('status', flat=True).distinct()
        self.fields['status'].choices = [('', 'All')] + [(e, e) for e in status_list]

    def clean_end_date(self):
        end_date = self.cleaned_data['end_date']
        start_date = self.cleaned_data['start_date']
        if start_date and end_date and end_date < start_date:
            raise forms.ValidationError("End date cannot be lesser than Start date")
        return self.cleaned_data['end_date']


class CustomExportMixin(import_export_admin.ExportMixin):
    export_template_name = 'admin/import_export/export.html'

    def get_export_formats(self):
        return [import_export.formats.base_formats.CSV, ]

    def get_export_queryset(self, request, cleaned_data=None):
        # can't manipulate/replace the queryset filters once added
        # http://stackoverflow.com/questions/5220433/how-to-edit-filters-list-of-a-queryset
        # http://stackoverflow.com/questions/4689566/django-remove-a-filter-condition-from-a-queryset
        # need to manipulate the request.GET multidict
        GET = request.GET.copy()
        export_params = [('start_date', 'timestamp__gte'), ('end_date', 'timestamp__lt'), ('status', 'status')]
        for field_name, param in export_params:
            if cleaned_data.get(field_name) is not None and GET.get(param):
                GET.pop(param)
        request = HttpRequest()
        request.GET = GET
        queryset = super(CustomExportMixin, self).get_export_queryset(request)

        if cleaned_data.get('start_date'):
            queryset = queryset.filter(timestamp__gte=cleaned_data['start_date'])

        if cleaned_data.get('end_date'):
            end_timestamp = datetime.datetime.combine(cleaned_data['end_date'], datetime.time.max)
            queryset = queryset.filter(timestamp__lte=end_timestamp)

        if cleaned_data.get('status'):
            queryset = queryset.filter(status=cleaned_data['status'])

        config = AccessLogConfiguration.get_solo()
        if cleaned_data.get('filter_bots'):
            config_agents = config.user_agent_bot_list.strip().split(',')
            user_agent_iregex = '(' + '|'.join([uas.strip() for uas in config_agents]) + ')'
            queryset = queryset.exclude(Q(user_agent__iregex=r'(%s)' % user_agent_iregex) | Q(user_agent='-'))

        if cleaned_data.get('filter_admin_ips'):
            start_date = datetime.datetime.today() - datetime.timedelta(days=int(cleaned_data['filter_admin_ips']))

            admin_ips = set(AccessLog.objects.filter(timestamp__gt=start_date, request__istartswith='/admin')
                            .values_list('remote_host', flat=True))
            config_excluded_ips = config.access_log_export_excluded_ips.strip().split(',')
            excluded_ips = [ip.strip() for ip in config_excluded_ips] + list(admin_ips)
            queryset = queryset.exclude(remote_host__in=excluded_ips)
        return queryset

    def export_action(self, request, *args, **kwargs):
        formats = self.get_export_formats()

        start_date = None
        if request.GET.get('timestamp__gte'):
            start_date = datetime.datetime.strptime(request.GET['timestamp__gte'].split(' ')[0], '%Y-%m-%d')

        end_data = None
        if request.GET.get('timestamp__lt'):
            end_data = datetime.datetime.strptime(request.GET['timestamp__lt'].split(' ')[0], '%Y-%m-%d')

        status = request.GET.get('status', None) or None

        initial_data = {
            'start_date': start_date,
            'end_date': end_data,
            'status': status
        }

        form = CustomExportForm(formats, request.POST or None, initial=initial_data)
        if form.is_valid():
            file_format = formats[int(form.cleaned_data['file_format'])]()

            queryset = self.get_export_queryset(request, cleaned_data=form.cleaned_data)
            export_data = self.get_export_data(file_format, queryset)
            content_type = 'application/octet-stream'
            # Django 1.7 uses the content_type kwarg instead of mimetype
            try:
                response = HttpResponse(export_data, content_type=content_type)
            except TypeError:
                response = HttpResponse(export_data, mimetype=content_type)
            response['Content-Disposition'] = 'attachment; filename=%s' % (
                self.get_export_filename(file_format),
            )
            return response

        context = {
            'form': form,
            'opts': self.model._meta,
        }

        if request.GET.get('q'):
            context['search_term'] = request.GET.get('q')

        return TemplateResponse(request, [self.export_template_name], context)


class HTTPMethodListFilter(SimpleListFilter):
    title = "Http Method"
    parameter_name = "http_method"

    def lookups(self, request, model_admin):
        return [('GET', 'GET'), ('POST', 'POST'), ('PUT', 'PUT'), ('HEAD', 'HEAD'),
                ('OPTIONS', 'OPTIONS'), ('DELETE', 'DELETE')]

    def queryset(self, request, qs):
        if self.value():
            qs = qs.filter(http_method__iexact=self.value())
        return qs


class AccessLogAdmin(CustomExportMixin, admin.ModelAdmin):
    list_display = ['timestamp', 'remote_host', 'status',
                    'http_method', 'request', 'referer',
                    'bytes_sent', 'http_version', 'user_agent_display',
                    'request_time', 'upstream_response_time', 'pipe']
    list_filter = ['timestamp', 'status', HTTPMethodListFilter]
    search_fields = ['timestamp', 'bytes_sent', 'referer',
                     'request', 'remote_host', 'status',
                     'user_agent']

    resource_class = AccessLogResource
    actions = None

    class Media:
        js = ('js/list_filter_collapse.js',)

    def has_change_permission(self, request, obj=None):
        return obj is None or request.method == 'GET'

    def change_view(self, request, object_id, form_url='', extra_context=None):
        extra_context = extra_context or {}
        extra_context['show_save_and_continue'] = False
        extra_context['show_save'] = False
        extra_context['save_as'] = False
        extra_context['add'] = False

        return super(AccessLogAdmin, self).change_view(request, object_id, extra_context=extra_context)

    def has_add_permission(self, request):
        return False

    def has_save_permission(self, *args):
        return False

    def has_delete_permission(self, *args):
        return False


admin.site.register(AccessLog, AccessLogAdmin)
admin.site.register(AccessLogConfiguration, SingletonModelAdmin)





# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AccessLog',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('timestamp', models.DateTimeField(db_index=True)),
                ('bytes_sent', models.CharField(max_length=500)),
                ('referer', models.TextField()),
                ('request', models.TextField()),
                ('http_method', models.CharField(max_length=500)),
                ('http_version', models.CharField(max_length=500)),
                ('remote_host', models.CharField(max_length=500)),
                ('remote_login', models.CharField(max_length=500)),
                ('remote_user', models.CharField(max_length=500)),
                ('status', models.CharField(max_length=500)),
                ('user_agent', models.CharField(max_length=500)),
                ('request_time', models.CharField(max_length=500)),
                ('upstream_response_time', models.CharField(max_length=500)),
                ('pipe', models.CharField(max_length=500)),
            ],
            options={
                'ordering': ('-timestamp',),
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-05-04 05:49
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('access_logs', '0002_accesslogconfiguration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='accesslogconfiguration',
            name='user_agent_bot_list',
            field=models.TextField(default=b'cloudflare, twiceler, yahooseeker, chtml, generic, heritrix, attentio, fast, mediapartners, python, experiment, fastmobilecrawl, curl, yahooysmcm, crawl, bingbot, bot, borg, google(^tv), yahoo, slurp, msnbot, msrbot, openbot, archiver, netresearch, lycos, scooter, altavista, teoma, gigabot, baiduspider, blitzbot, oegp, charlotte, furlbot, http%20client, polybot, htdig, ichiro, mogimogi, larbin, pompos, scrubby, searchsight, seekbot, semanticdiscovery, silk, snappy, speedy, spider, voila, vortex, voyager, zao, zeal, fast\\-webcrawler, converacrawler, dataparksearch, findlinks, crawler, Netvibes, Sogou Pic Spider, ICC\\-Crawler, Innovazion Crawler, Daumoa, EtaoSpider, A6\\-Indexer, YisouSpider, Riddler, DBot, wsr\\-agent, Xenu, SeznamBot, PaperLiBot, SputnikBot, CCBot, ProoXiBot, Scrapy, Genieo, Screaming Frog, YahooCacheSystem, CiBra, Nutch, holmes, WebThumbnail, ^voyager, heritrix, scraper, favicon, Google.*/\\+/web/snippet, Icarus6j, PagePeeker, ^vortex, Sogou, ^Java/, BlogBridge, ZooShot, indexer, GomezAgent, ^JNLP/, ^NING, WinHTTP, TLSProber, Squrl Java, NewsGator, Google-HTTP-Java-Client, Reaper, WhatWeb, crawl, facebookexternalhit, Python-urllib, IlTrovatore-Setaccio, AppEngine-Google, GoogleStackdriverMonitoring, InternetArchive, WordPress, Retreiver'),
        ),
    ]



import re
import dateutil.parser
from celery.schedules import crontab
from celery.task import periodic_task
from django.conf import settings

from .models import AccessLog

regex = re.compile(r"(?P<remote_host>\S*)\s-\s(?P<remote_user>\S*)\s\[(?P<timestamp>.*?)\]\s{1,2}\"(?P<http_method>\S*)\s*(?P<request>\S*)\s*(HTTP\/)*(?P<http_version>.*?)\"\s(?P<status>\d{3})\s(?P<bytes_sent>\S*)\s\"(?P<referer>[^\"]*)\"\s\"(?P<user_agent>[^\"]*)\"\s(?P<request_time>\S*)\s(?P<upstream_response_time>\S*)\s(?P<pipe>\S*)")

# Run before logrorate rotates the file
@periodic_task(run_every=crontab(minute=23, hour=6))
def parse_access_log():
    parse_access_log_file(settings.ACCESS_LOG_PATH)


def parse_access_log_file(file_path):
    errors = {}
    with open(file_path, 'r') as log_file:
        for line in log_file:
            try:
                r = regex.search(line.strip())
                if r is None:
                    continue
                data = r.groupdict()
                if re.search("(\.gif|\.jpg|\.png|\.css|\.js|\.ico|\.txt)/?$", data['request']):
                    continue
                if any(map(lambda x: len(data.get(x)) > 500, ['http_method', 'http_version', 'user_agent'])):
                    continue
                data['timestamp'] = dateutil.parser.parse(data['timestamp'], fuzzy=True)
                AccessLog.objects.get_or_create(**data)
            # pylint: disable=W0703
            except Exception as e:
                errors[line] = e

    if errors:
        raise Exception(errors)


import os
from setuptools import setup

__version__ = "0.1.11"

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django_access_logs',
    version=__version__,
    packages=['access_logs', 'access_logs.migrations'],
    include_package_data=True,
    license='Apache License, Version 2.0',
    description='A simple module to record server access logs in DB and export them',
    long_description="A simple reusable Django app to record parsed server access logs "
                     "in a database and export them as CSV on demand",
    url='https://github.com/TriplePoint-Software/django_access_logs',
    author='Jai',
    author_email='jaivikram.verma@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'celery>=3.1.18',
        'django-celery>=3.1.16',
        'django-import-export>=0.2.7',
        'django-solo>=1.1.2',
        'django>=1.10,<1.11',
        'python-dateutil>=2.4.2',
        'ua-parser>=0.3.6',
    ],
)
